#ifndef GR_ACL_H
#define GR_ACL_H
#endif
#include <linux/grsecurity.h>
#include <linux/grdefs.h>

/* * * * * * * * * * * * * * * * * * * * *
 * GrSecurity ACL System
 * Main header file
 * Purpose: define most gracl data structures 
 * * * * * * * * * * * * * * * * * * * * */

/* Major status information */

#define GR_VERSION  "grsecurity 2.0"

enum
{

	SHUTDOWN   = 0,
	ENABLE     = 1,
	CHCAPS     = 2,
	AUTH       = 3,
	GOD        = 4,
	RELOAD     = 5,
	
	/* Number of capabilities 
	 * see include/linux/capability.h */
	CAP_COUNT = 30
};

/* Internal status definitions
 * gr_set_proc_label() in kernel/gracl.c
 * */
enum
{
	GR_SET_INHERIT,
	GR_SET_AUTH,   
	GR_SET_PROC
};

/* Password setup definitions
 * kernel/grhash.c */
enum
{
	GR_PW_LEN    = 129,
	GR_SHA_SIZE  = 20,
	GR_EXTRA_LEN = 1025
};

/* Begin Data Structures */



struct admin_pw
{
	char 				pw[GR_PW_LEN];
	char 				extra[GR_EXTRA_LEN];
	__u8 				sum[GR_SHA_SIZE];  /* 160-bit SHA hash of the password*/
	__u8 				mode; /* On/Off/Restart*/
};

struct file_label
{
	ino_t 				inode;
	kdev_t 				device;
	__u32 				mode;
};

/* Begin TBD */
#if 0
struct socket_label 
{
	unsigned int 			port;
	__u8 				mode:1; /* Connect or Bind */
};

struct id_label 
{
	__u32 				id;
	__u8 				id_type:1; /* UID or GID */
	__u8 				mode; /* execlog/ptrace/etc.*/
						      
	struct mac_socket *		sockets;
};
#endif
/* end TBD */

struct file_db 
{
	struct file_label *         	file_hash;
	unsigned long             	file_size;
	unsigned long			file_used;
};


struct proc_subject_label
{
	ino_t                     	inode;
	kdev_t                    	device;
	__u16                      	mode; 
	kernel_cap_t              	cap_raise; 
	kernel_cap_t			cap_lower;
	kernel_cap_t			cap_raise_inherit;
	kernel_cap_t			cap_lower_inherit;
	/* Keeps track of the capabilities to raise and lower. 
	 * Have to store it this way because users my change the default
	 * capabilities, and we want to change with them(but still
	 * keep our custom raised/lowered caps) */
	unsigned long 			pos; 
	/* .. used for process object
	hashing. Position (i.e.
	subject acl number 1..n) */
};


struct proc_object_label
{
	ino_t                     	inode;
	kdev_t                    	device; 
	__u16                     	mode; 
	unsigned long			parent_pos;
};

struct name_entry
{
	ino_t inode;
	kdev_t device;
	__u8 mode;
	char * name;
};
	
struct proc_subject_db
{
	struct proc_subject_label  *	proc_hash;
	unsigned long            	proc_size;
	unsigned long			proc_used;
};

struct proc_object_db
{
	struct proc_object_label  * 	proc_hash;
	unsigned long            	proc_size;
	unsigned long			proc_used;
};

struct name_db
{
	struct name_entry *		 name_hash;
	unsigned long			 name_size;
	unsigned long			 name_used;
};

struct temp_file_label
{
	ino_t 				inode;
	kdev_t 				device;
	__u32 				mode;
};

struct proc_info
{
	ino_t				inode;
	kdev_t				device;
	unsigned long			pos;
};

struct temp_socket_label
{
	unsigned int 			port;
	__u8 				mode;
};

struct label_info 
{
	ino_t inode;
	kdev_t device;
};




struct cap_entry 
{
	char * 				capname;
	kernel_cap_t 			capval;
};

struct capset
{
	struct cap_entry 		table[CAP_COUNT];
};

/* End Data Structures Section */

static __inline__ unsigned long  fhash(const ino_t ino,const kdev_t dev,
		const unsigned long sz)
{
return  (ino ^ (dev * 797) )  % sz;
}

static __inline__ unsigned long phash(const ino_t ino, const kdev_t dev, 
		const unsigned long parent_pos, const unsigned long sz)
{
	/* Okay...parent_pos is the subject acl # (0..n-1 for n subject acls)
	 * We reuse fhash here for the parent_pos portion of phash() for
	 * simplicity's sake */
	return ( fhash(ino,dev,~0UL) ^ fhash((ino_t)(parent_pos),0,~0UL) ) 
		% sz;
}

static __inline__ unsigned long nhash(const char * name, const unsigned long sz)
{
	return full_name_hash(name,strlen(name)) % sz;
}
			
			




/* * * * * * * * * * * * * * * * * * * *
 * From this point on, the header
 * file for the SHA-1 algorithm
 * is defined. This is slightly
 * adapted from gnupg. All errors
 * are mine, all features are theirs:).
 * That is why there is a style
 * difference. See kernel/grhash.c 
 * * * * * * * * * * * * * * * * * * * */






/* Begin SHA hash section */
#define GR_SHA1HANDSOFF /* Copies data before messing with it. */

struct gr_SHA1_CTX {
	unsigned long 			state[5];
	unsigned long 			count[2];
	unsigned char 			buffer[64];
};

#define rol(value, bits) (((value) << (bits)) | ((value) >> (32 - (bits))))

/* blk0() and blk() perform the initial expand. */
/* I got the idea of expanding during the round function from SSLeay */
#ifdef __LITTLE_ENDIAN
#define blk0(i) (block->l[i] = (rol(block->l[i],24)&0xFF00FF00) \
		|(rol(block->l[i],8)&0x00FF00FF))
#else
#define blk0(i) block->l[i]
#endif /* __LITTLE_ENDIAN */
#define blk(i) (block->l[i&15] = rol(block->l[(i+13)&15]^block->l[(i+8)&15] \
			^block->l[(i+2)&15]^block->l[i&15],1))

	/* (R0+R1), R2, R3, R4 are the different operations used in SHA1 */
#define R0(v,w,x,y,z,i) z+=((w&(x^y))^y)+blk0(i)+0x5A827999+rol(v,5);w=rol(w,30);
#define R1(v,w,x,y,z,i) z+=((w&(x^y))^y)+blk(i)+0x5A827999+rol(v,5);w=rol(w,30);
#define R2(v,w,x,y,z,i) z+=(w^x^y)+blk(i)+0x6ED9EBA1+rol(v,5);w=rol(w,30);
#define R3(v,w,x,y,z,i) z+=(((w|x)&y)|(w&x))+blk(i)+0x8F1BBCDC+rol(v,5);w=rol(w,30);
#define R4(v,w,x,y,z,i) z+=(w^x^y)+blk(i)+0xCA62C1D6+rol(v,5);w=rol(w,30);
/* End SHA hash section */		       
